from typing import Any

from pydantic import BaseModel

from icij_common.pydantic_utils import icij_config
from icij_worker import TaskState


class TaskSearch(BaseModel):
    model_config = icij_config()

    name: str | None = None
    status: list[TaskState] | TaskState | None = None


class TaskCreationQuery(BaseModel):
    model_config = icij_config()

    name: str
    args: dict[str, Any]



import logging
from http.client import HTTPException

from fastapi import APIRouter
from starlette.responses import Response
from starlette.status import HTTP_204_NO_CONTENT

from icij_common.logging_utils import TRACE, log_elapsed_time_cm
from icij_worker import Task
from icij_worker.exceptions import UnknownTask
from icij_worker.http_.constants import TASKS_TAG
from icij_worker.http_.dependencies import lifespan_task_manager
from icij_worker.http_.objects import TaskCreationQuery, TaskSearch
from icij_worker.objects import ErrorEvent

logger = logging.getLogger(__name__)


def tasks_router(prefix: str = "/api") -> APIRouter:
    router = APIRouter(prefix=prefix, tags=[TASKS_TAG])

    @router.put("/tasks/{task_id}")
    async def _create_task(task_id: str, task: TaskCreationQuery) -> str:
        task = Task.create(task_id=task_id, task_name=task.name, args=task.args)
        task_manager = lifespan_task_manager()
        is_new = await task_manager.save_task(task)
        if not is_new:
            logger.debug('Task(id="%s") already exists, skipping...', task_id)
            return Response(task.id, status_code=200)
        logger.debug('Task(id="%s") created, queuing...', task_id)
        await task_manager.enqueue(task)
        logger.info('Task(id="%s") queued...', task_id)
        return Response(task.id, status_code=201)

    async def _get_task_(task_id: str) -> Task:
        task_manager = lifespan_task_manager()
        try:
            with log_elapsed_time_cm(
                logger, logging.INFO, "retrieved task in {elapsed_time} !"
            ):
                task = await task_manager.get_task(task_id=task_id)
        except UnknownTask as e:
            raise HTTPException(status_code=404, detail=e.args[0]) from e
        return task

    @router.get("/tasks/{task_id}")
    async def _get_task(task_id: str) -> Task:
        return await _get_task_(task_id)

    @router.get("/tasks/{task_id}/state", response_model=str)
    async def _get_task_state(task_id: str) -> str:
        state = (await _get_task_(task_id)).state.value
        return Response(content=state, media_type="text/plain")

    @router.post("/tasks/{task_id}/cancel", responses={204: {"model": None}})
    async def _cancel_task(
        task_id: str,
        requeue: bool = False,
    ) -> None:  # noqa: FBT001, FBT002
        task_manager = lifespan_task_manager()
        try:
            await task_manager.cancel(task_id=task_id, requeue=requeue)
        except UnknownTask as e:
            raise HTTPException(status_code=404, detail=e.args[0]) from e
        return Response(status_code=HTTP_204_NO_CONTENT)

    @router.get("/tasks/{task_id}/result", response_model=object)
    async def _get_task_result(task_id: str) -> object:
        task_manager = lifespan_task_manager()
        try:
            result = await task_manager.get_task_result(task_id=task_id)
        except UnknownTask as e:
            raise HTTPException(status_code=404, detail=e.args[0]) from e
        return result.result

    @router.get("/tasks/{task_id}/errors")
    async def _get_task_errors(task_id: str) -> list[ErrorEvent]:
        task_manager = lifespan_task_manager()
        try:
            errors = await task_manager.get_task_errors(task_id=task_id)
        except UnknownTask as e:
            raise HTTPException(status_code=404, detail=e.args[0]) from e
        return errors

    @router.post("/tasks", response_model=list[Task])
    async def _search_tasks(search: TaskSearch) -> list[Task]:
        task_manager = lifespan_task_manager()
        with log_elapsed_time_cm(logger, TRACE, "searched tasks in {elapsed_time} !"):
            tasks = await task_manager.get_tasks(
                group=None, task_type=search.name, status=search.status
            )
        return tasks

    return router



from __future__ import annotations

from abc import ABC
from pathlib import Path
from typing import ClassVar

from pydantic import Field
from pydantic_settings import SettingsConfigDict

from icij_common.pydantic_utils import icij_config, merge_configs
from icij_common.registrable import RegistrableSettings
from icij_worker import AsyncBackend


class WorkerConfig(RegistrableSettings, ABC):
    model_config = merge_configs(
        icij_config(),
        SettingsConfigDict(env_prefix="ICIJ_WORKER_", env_nested_delimiter="_"),
    )

    registry_key: ClassVar[str] = Field(frozen=True, default="type")
    type: AsyncBackend = Field(frozen=True)

    # TODO: is app_dependencies_path better ?
    app_bootstrap_config_path: Path | None = None
    inactive_after_s: float | None = None
    log_level: str = "INFO"
    task_queue_poll_interval_s: float = 1.0



from icij_worker.typing_ import RateProgress, RawProgress


def to_raw_progress(progress: RateProgress, max_progress: int) -> RawProgress:
    if not max_progress > 0:
        raise ValueError("max_progress must be > 0")

    async def raw(p: int):
        await progress(p / max_progress)

    return raw


def to_scaled_progress(progress: RateProgress, *, start: float = 0.0, end: float = 1.0):
    if not 0 <= start < end:
        raise ValueError("start must be [0, end[")
    if not start < end <= 1.0:
        raise ValueError("end must be ]start, 1.0]")

    async def _scaled(p: float):
        await progress(start + p * (end - start))

    return _scaled



from typing import Annotated, ClassVar, Union

from pydantic import Discriminator, Field, Tag
from pydantic_settings import BaseSettings, SettingsConfigDict

import icij_worker
from icij_common.pydantic_utils import (
    icij_config,
    make_enum_discriminator,
    merge_configs,
)
from icij_worker import AsyncBackend, TaskManager
from icij_worker.utils.logging_ import LogWithWorkerIDMixin

_PLACEHOLDER = ""
_ALL_LOGGERS = [icij_worker.__name__]


def _get_available_tm_configs() -> tuple[Annotated, ...]:
    try:
        from icij_worker.task_manager.neo4j_ import Neo4JTaskManagerConfig
    except ImportError:
        Neo4JTaskManagerConfig = None
    try:
        from icij_worker.task_manager.amqp import AMQPTaskManagerConfig
    except ImportError:
        AMQPTaskManagerConfig = None

    configs = [AMQPTaskManagerConfig, Neo4JTaskManagerConfig]
    configs = tuple(
        Annotated[c, Tag(AsyncBackend(c.registered_name))]
        for c in configs
        if c is not None
    )
    return configs


TM = Union[_get_available_tm_configs()]  # pylint: disable=invalid-name


class HttpServiceConfig(BaseSettings, LogWithWorkerIDMixin):
    model_config = merge_configs(
        icij_config(),
        SettingsConfigDict(
            env_prefix=_PLACEHOLDER, env_nested_delimiter="__", validate_default=True
        ),
    )

    app_title: str = "HTTP service"
    app_path: ClassVar[str] = Field(default=_PLACEHOLDER, frozen=True)
    loggers: ClassVar[list[str]] = Field(
        _ALL_LOGGERS + ["uvicorn", "__main__"], frozen=True
    )
    host: str = "localhost"
    log_level: str = Field(default="INFO")
    n_workers: int = 1
    port: int = 8080

    task_manager: TM = Field(
        discriminator=Discriminator(make_enum_discriminator("backend", AsyncBackend))
    )

    def to_task_manager(self) -> TM:
        return TaskManager.from_config(self.task_manager)



import logging
from contextlib import AbstractAsyncContextManager, asynccontextmanager
from typing import cast

from fastapi import FastAPI

from icij_worker import AMQPTaskManager
from icij_worker.http_.config import HttpServiceConfig
from icij_worker.typing_ import Dependency
from icij_worker.utils.dependencies import DependencyInjectionError, run_deps

_TASK_MANAGER: AMQPTaskManager | None = None
_HTTP_SERVICE_CONFIG: HttpServiceConfig | None = None

logger = logging.getLogger(__name__)


async def http_service_config_setup(config: HttpServiceConfig | None, **_) -> None:
    global _HTTP_SERVICE_CONFIG
    _HTTP_SERVICE_CONFIG = config
    logger.info(
        "HTTP Service config: %s\n", _HTTP_SERVICE_CONFIG.model_dump_json(indent=2)
    )


def lifespan_http_service_config() -> HttpServiceConfig:
    if _HTTP_SERVICE_CONFIG is None:
        raise DependencyInjectionError("http service config")
    return cast(HttpServiceConfig, _HTTP_SERVICE_CONFIG)


async def loggers_setup(**_) -> None:
    config = cast(HttpServiceConfig, lifespan_http_service_config())
    config.setup_loggers()


async def task_manager_setup(**_) -> None:
    global _TASK_MANAGER
    config = cast(HttpServiceConfig, lifespan_http_service_config())
    _TASK_MANAGER = config.to_task_manager()
    await _TASK_MANAGER.__aenter__()  # pylint: disable=unnecessary-dunder-call


async def task_manager_teardown(exc_type, exc_val, exc_tb) -> None:  # noqa: ANN001
    global _TASK_MANAGER
    await _TASK_MANAGER.__aexit__(exc_type, exc_val, exc_tb)
    _TASK_MANAGER = None


def lifespan_task_manager() -> AMQPTaskManager:
    if _TASK_MANAGER is None:
        raise DependencyInjectionError("task manager")
    return cast(AMQPTaskManager, _TASK_MANAGER)


BASE_DEPENDENCIES = [
    ("http service config", http_service_config_setup, None),
    ("loggers", loggers_setup, None),
    ("task manager", task_manager_setup, task_manager_teardown),
]


@asynccontextmanager
async def run_http_service_deps(
    app: FastAPI, config: HttpServiceConfig, dependencies: list[Dependency]
) -> AbstractAsyncContextManager:
    app.state.config = config
    async with run_deps(dependencies=dependencies, ctx=app.title, config=config):
        yield



import logging

import sys
from pydantic.fields import FieldInfo

import icij_worker
from icij_common.logging_utils import (
    DATE_FMT,
    STREAM_HANDLER_FMT,
    STREAM_HANDLER_FMT_WITH_WORKER_ID,
    WorkerIdFilter,
)
from icij_common.pydantic_utils import get_field_default_value


class LogWithWorkerIDMixin:
    def setup_loggers(self, worker_id: str | None = None):
        # Ugly work around the Pydantic V1 limitations...
        all_loggers = self.loggers
        if isinstance(all_loggers, FieldInfo):
            all_loggers = get_field_default_value(all_loggers)
        all_loggers.append(icij_worker.__name__)
        loggers = sorted(set(all_loggers))
        log_level = self.log_level
        if isinstance(log_level, FieldInfo):
            log_level = get_field_default_value(log_level)
        force_warning = getattr(self, "force_warning_loggers", [])
        if isinstance(force_warning, FieldInfo):
            force_warning = get_field_default_value(force_warning)
        force_warning = set(force_warning)
        worker_id_filter = None
        if worker_id is not None:
            worker_id_filter = WorkerIdFilter(worker_id)
        handlers = self._handlers(worker_id_filter, log_level)
        for logger in loggers:
            logger = logging.getLogger(logger)
            level = getattr(logging, log_level)
            if logger.name in force_warning:
                level = max(logging.WARNING, level)
            logger.setLevel(level)
            logger.handlers = []
            for handler in handlers:
                logger.addHandler(handler)

    def _handlers(
        self, worker_id_filter: logging.Filter | None, log_level: int
    ) -> list[logging.Handler]:
        stream_handler = logging.StreamHandler(sys.stderr)
        if worker_id_filter is not None:
            fmt = STREAM_HANDLER_FMT_WITH_WORKER_ID
        else:
            fmt = STREAM_HANDLER_FMT
        log_in_json = getattr(self, "log_in_json", False)
        if isinstance(log_in_json, FieldInfo):
            log_in_json = get_field_default_value(log_in_json)
        if log_in_json:
            # TO be installed and required on child libs
            # TODO: add it as an extra of this lib
            from pythonjsonlogger.jsonlogger import JsonFormatter

            fmt = JsonFormatter(fmt, DATE_FMT)
        else:
            fmt = logging.Formatter(fmt, DATE_FMT)
        stream_handler.setFormatter(fmt)
        handlers = [stream_handler]
        for handler in handlers:
            if worker_id_filter is not None:
                handler.addFilter(worker_id_filter)
            handler.setLevel(log_level)
        return handlers
